#ifndef BisqPAPUinternals_hh
#define BisqPAPUinternals_hh

class sndchan
{
public:
	bool enabled;
	SBYTE outputvalue;
public:
	sndchan() : enabled(true), outputvalue(0) { }
	inline void enable() { enabled = true; }
};

#define PAPUTICK_60HZ(val) (!((val)&3))    // 00 but no 01 02 03   (1/4)
#define PAPUTICK_120HZ(val) (!((val)&1))   // 00 02 but no 01 03   (1/2)
#define PAPUTICK_240HZ(val) true           // 00 01 02 03          (1/1)

class waveunit : public sndchan
{
	PAIR wavelen;
	WORD wavepos;
public:
	waveunit() : wavepos(0)
	{
		wavelen.W = 0;
	}
	void waveunit_write(fBYTE ind, fBYTE value)
	{
		if(ind == 2)wavelen.B.l = value;
		else if(ind == 3)wavelen.B.h = value&7;
	//	fprintf(stderr, "Got period %u\n", wavelen.W);
	}
	inline WORD getwavelen() const { return wavelen.W; }
	bool setwavelen(fWORD len)
	{
		register bool carry;
		if(len > 0x7FF)
		{
			len &= 0x7FF;
			carry = true;
		}
		else
			carry = false;
		
		wavelen.W = len;
		return carry;
	}
	
	virtual void wavetick() = 0;
	
	// This requires CPU tick
	void waveunit_tick()
	{
		if(++wavepos >= wavelen.W)
		{
			wavepos = 0;
			wavetick();
		}
	}
};

class lengthunit
{
	unsigned lengthcounter;
	BYTE bmask;
	bool lengthdisabled;
public:
	lengthunit(BYTE mask = 5) : lengthcounter(0), bmask(mask), lengthdisabled(true)
	{
	}
	void lengthunit_write(fBYTE ind, fBYTE value)
    {
    	if(ind == 3)
    	{
            static const BYTE lengthtrans[32]=
            {5,127,10,1,20, 2,40, 3,80, 4,30, 5,7, 6,14, 7,
             6,  8,12,9,24,10,48,11,96,12,36,13,8,14,16,15};
            // Ks papu.txt jos tuon taulukon sislt kiinnostaa
            
    		lengthcounter = lengthtrans[(value >> 3) & 31];
    	}
    	if(ind == 0)
    	{
    		lengthdisabled = (value & bmask);
    		//fprintf(stderr, "lengthunit: disabled set to %u\n", lengthdisabled);
    	}
    }
	inline void lengthunit_reset()
	{
		lengthcounter = 0;
	}
	inline bool lengthunit_done() const
	{
		//if(!lengthdisabled)
		//	fprintf(stderr, "lengthunit: counter=%u\n", lengthcounter);
		return !lengthcounter;
	}
	void lengthunit_tick(BYTE ticktype)
	{
		if(lengthdisabled)return;		
		if(PAPUTICK_60HZ(ticktype))
		{
			if(lengthcounter > 0)--lengthcounter;
		}
	}
};

class decayunit : public lengthunit
{
	BYTE decaycounter;
	BYTE decayrate;
	BYTE decaypos;
	BYTE decaymode;
public:
	decayunit() : decaycounter(0), decayrate(0), decaypos(0), decaymode(0x10)
	{
	}
	void decayunit_write(fBYTE ind, fBYTE value)
	{
		lengthunit_write(ind, value);
		
		if(ind == 0)
		{
			decayrate = value & 15;
			decaymode = (value>>4) & 3;
		}
		else if(ind == 3)
			decaycounter = 15;
	}
	inline unsigned getvol() const
	{
		return (decaymode & 1) ? decayrate : decaycounter;
	}
	void decayunit_tick(BYTE ticktype)
	{
		lengthunit_tick(ticktype);
		
		if(PAPUTICK_240HZ(ticktype))
		{
			if(!decaypos)
			{
				decaypos = decayrate;
				if(decaycounter>0)
					--decaycounter;
				else if(decaymode & 2)
					decaycounter = 15;
			}
			else
				--decaypos;
		}
	}
};

template<class T>
class papuhzticker : public T
{
	unsigned hzcounter, countermax;
	BYTE hzpos;
public:
	papuhzticker(BYTE n=0) : T(n), hzcounter(0), hzpos(0)
	{
		countermax = Clock->ClockHz / 240;
	}
	inline void tick()
	{
		waveunit_tick();
		if(++hzcounter >= countermax)
		{
			hzcounter = 0;
			hztick(++hzpos);
		}
	}
};

class triangleunit : public waveunit, public lengthunit
{
	unsigned trianglecounter;
	unsigned linearcounter;
	enum {triload, tricount} trimode;
	unsigned triloadvalue;
	bool oldbit;
public:
	triangleunit(BYTE) :lengthunit(7),  trianglecounter(0), linearcounter(0), trimode(triload), triloadvalue(0), oldbit(false)
	{
	}
	void write(fBYTE ind, fBYTE value)
	{
		//fprintf(stderr, "triangleunit::write(%u, 0x%02X);\n", ind, value);
		waveunit_write(ind, value);
		lengthunit_write(ind, value);
		
		if(ind == 3)
		{
			trimode = (oldbit || !linearcounter) ? triload : tricount;
			
		}
		else if(ind == 0)
		{
			triloadvalue = value & 0x7F;
			bool newbit = value & 0x80;
			
			if(!oldbit)
			{
				trimode = tricount;
				linearcounter = triloadvalue;
			}
			else if(!newbit && !linearcounter)
				trimode = tricount;
			
			newbit = oldbit;
		}
	}
	void hztick(BYTE ticktype)
	{
		lengthunit_tick(ticktype);
		
		if(PAPUTICK_240HZ(ticktype))
		{
			fprintf(stderr, "triangle tick, linearcounter=%u\n", linearcounter);		
			if(trimode==tricount && linearcounter > 0)
				--linearcounter;
		}
	}
	virtual void wavetick()
	{
		if(!silent())
		{
            static const BYTE triangledata[32]=
            {13,14,15,15,14,13,12,11,10, 9, 8, 7, 6, 5, 4, 3,
              2, 1, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12};

			outputvalue = triangledata[trianglecounter = (trianglecounter+1)&31] * 4;
		//	fprintf(stderr, "TRIANGLE TICK - outputvalue = %d\n", outputvalue);
		}
	}
	inline bool silent() const
	{
		if(!enabled)return true;
		if(lengthunit_done())return true;
		if(!linearcounter)return true;
		if(getwavelen() < 8)return true;
		return false;
	}
};

class squaregenunit : public waveunit, public decayunit
{
	BYTE sweepdata;
	BYTE sweeppos;
	bool jammed;
	BYTE dutytype;
	BYTE dutycounter;
	BYTE decr;
public:
	squaregenunit(BYTE d) : sweepdata(0x00), sweeppos(0), jammed(false), dutytype(0), decr(d)
	{ 
	}
	void write(fBYTE ind, fBYTE value)
	{
		waveunit_write(ind, value);
		decayunit_write(ind, value);
		
		//fprintf(stderr, "squaregenunit::write(%u, 0x%02X);\n", ind, value);
		jammed = false;
		if(ind == 1)
			sweepdata = value;
		else if(ind == 0)
		{
			if((value >> 6) != dutytype)
				fprintf(stderr, "squaregenunit: NEW DUTYTYPE - %03u\n", value >> 6);
			dutytype = value >> 6;
		}
	}
	void hztick(BYTE ticktype)
	{
		if(PAPUTICK_120HZ(ticktype))
		{
			if(!sweeppos)
			{
				sweeppos = ((sweepdata >> 4) & 7) + 1;
				if(sweepdata & 0x80)
				{
					unsigned len = getwavelen();
					if(sweepdata & 8)
						len -= (len >> (sweepdata & 7)) + decr;
					else
						len += (len >> (sweepdata & 7));
					if(setwavelen(len) || len < 8)
						jammed = true;
				}
			}
			else
				--sweeppos;
		}
	}
	inline bool silent() const
	{
		if(!enabled)return true;
		if(jammed)  return true;
		if(lengthunit_done())return true;
		if(getwavelen() < 8)return true;
		return false;
	}
	inline int calcduty()
	{
		static const BYTE dutyvalues[4] = {2,4,8,12};
		return (++dutycounter&15) < dutyvalues[dutytype] ? -1 : 1;
	}
	virtual void wavetick()
	{
		if(!silent())
		{
		//	fprintf(stderr, "Square tick - volume = %d\n", getvol());
			outputvalue = calcduty() * getvol();
		}
	}
};

class randgenunit : public sndchan, public decayunit
{
	WORD wlen;
	bool randmode;  // false = 32767-bit, true = 93-bit
	WORD shiftreg;
	bool shifttick;
	unsigned wavepos;
	mutable bool booli;
public:
	randgenunit(BYTE): wlen(0), shiftreg(1), shifttick(false), wavepos(0)
	{
	}
	void write(fBYTE ind, fBYTE value)
	{
		fprintf(stderr, "randgenunit::write(%u, 0x%02X);\n", ind, value);
    	decayunit_write(ind, value);
    	
		if(ind == 2)
		{
            static const WORD wlenconvdata[16] =
            {0x002, 0x004, 0x008, 0x010,
             0x020, 0x030, 0x040, 0x050,
             0x065, 0x07F, 0x0BE, 0x0FE,
             0x17D, 0x1FC, 0x3F9, 0x7F2};
			wlen = wlenconvdata[value & 15];
			randmode = value >> 7;
		}
		booli = true;
	}
	inline bool silent() const
	{
		if(!enabled)return true;
		if(lengthunit_done()) { if(booli){fprintf(stderr, "RND is silent\n");booli=false;} return true; }
		if(booli){fprintf(stderr, "RND is noiseful\n");booli=false;}
		return false;
	}
	 void hztick(BYTE ticktype)
	{
		lengthunit_tick(ticktype);
	}
	inline void waveunit_tick() const { }
	void tick()
	{
    	// This requires CPU tick
   		if(++wavepos >= wlen)
   		{
   			wavepos = 0;
   			
    		if(shifttick && !silent())
    		{
       			BYTE result = ((shiftreg >> 14) ^ (shiftreg >> (randmode ? 8 : 13)) ^ 1) & 1;
        		outputvalue = result * getvol();
        		
	    		shiftreg = (shiftreg << 1) | result;
    			shifttick = false;
    		}
    		else
    			shifttick = true;
		}
	}
};

class sq1gen : public papuhzticker<squaregenunit>
{
public:
	sq1gen()
	{
	}
	inline void disable()
	{
		enabled = false;
		lengthunit_reset();
	}
};

class sq2gen : public papuhzticker<squaregenunit>
{
public:
	sq2gen() : papuhzticker<squaregenunit> (1)
	{
	}
	inline void disable()
	{
		enabled = false;
		lengthunit_reset();
	}
};

class trigen : public papuhzticker<triangleunit>
{
public:
	trigen()
	{
	}
	inline void disable()
	{
		enabled = false;
		lengthunit_reset();
	}
};

class rndgen : public papuhzticker<randgenunit>
{
public:
	inline void disable()
	{
		enabled = false;
		lengthunit_reset();
	}
	virtual bool silent() const
	{
		if(!enabled)return true;
		if(lengthunit_done())return true;
		
		return false;
	}
};

class dmcgen : public sndchan
{
public:
	void write(fBYTE ind, fBYTE value)
	{
	}
	inline void disable()
	{
		enabled = false;
	}
	virtual bool silent() const
	{
		if(!enabled)return true;
		return false;
	}
};

#endif
