#include "ppu.hh"
#include "clock.h"

void bisqppu::writemem(fWORD addr, fBYTE value)
{
	addr &= 0x3FFF;
	if(addr < 0x2000) { if(!vrom)vpage[addr/0x400][addr&0x3FF] = value; return; }
	if(addr < 0x3F00)
	{
		addr &= 0x1FFF;
		unsigned ind = addr / 0x400;
		namtab[ind][addr & 0x3FF] = value;
	}
	addr &= 0x0F;
	palette[addr | 0x10] = palette[addr] = value & 0x3F;
}

BYTE bisqppu::readmem(fWORD addr) const
{
	addr &= 0x3FFF;
	if(addr < 0x2000) return vpage[addr/0x400][addr&0x3FF];
	if(addr < 0x3F00)
	{
		addr &= 0x1FFF;
		unsigned ind = addr / 0x400;
		return namtab[ind][addr & 0x3FF];
	}
	return palette[addr & 0x1F];
}

bool bisqppu::write(const fWORD addr, fBYTE value)
{
	switch(addr)
	{
		case 0x2000: reg2000 = value; fprintf(stderr, "<NMI %sabled>", (value&80) ? "en":"dis"); fflush(stderr); return true;
		case 0x2001: reg2001 = value; return true;
		case 0x2003: reg2003 = value; return true;
		case 0x2004:
			spr_ram[reg2003++] = value;
			return true;
		case 0x2006:
			memptr = (memptr<<8) | value;
			return true;
		case 0x2007:
			writemem(memptr, value);
			memptr += (reg2000&4) ? 32 : 1;
			return true;
	}
	return false;
}

bool bisqppu::read(const fWORD addr, BYTE &value)
{
	switch(addr)
	{
		case 0x2002:
			if(vblank) { value |= 0x80; vblank = false; }
			return true;
		case 0x2007:
			value = readmem(memptr);
			memptr += (reg2000&4) ? 32 : 1;
			return true;
	}
	return false;
}

/* We do it *exactly* like NES, no matter how slow it is */
InterruptType bisqppu::tick()
{
	InterruptType ret = INT_NONE;
	if(scanlid < 240)
	{
		if(tickpos < 1360) // 170*2
		{
			if(!(tickpos & 3)) // kukin nist sy nelj ticki
        	{
        		if(tickpos < 128*8)
        		{
        			// Draw background 8 pix
        		}
        		else
        		{
        			// Fetch sprite for next scanline
        		}
        	}
        	++tickpos;
        }
        else if(tickpos == 1364-1)
        {
        	tickpos = 0;
        	++scanlid;
        }
        else
        {
        	++tickpos;
        	// 4 cycles of idle here (HBlank).
        }
	}
	else if(scanlid == 240) // One idle scanline is true NES
	{
		if(tickpos == 1364-1)
		{
			tickpos = 0;
			++scanlid;
			
			fprintf(stderr, "\33[0;31mVB\33[m");
			fflush(stderr);
			vblank = true;
			if(reg2000 & 0x80)
			{
				fprintf(stderr, "\33[0;1;31mN\33[m");
				fflush(stderr);
				ret = INT_NMI;
			}
		}
		else
			++tickpos;
	}
	else if(scanlid == Clock->ScanlineCount-1)
	{
		if(tickpos == (shortline ? 1360 : 1364)-1)
		{
			tickpos = 0;
			scanlid = 0;
			shortline = !shortline;
			vblank = false;
		}
		else
			++tickpos;
	}
	else
	{
		if(tickpos == 1364-1)
		{
			tickpos = 0;
			++scanlid;
		}
		else
		{
			// Doing a part of VBlank
			++tickpos;
		}
	}
	return ret;
}
